#include <jni.h>
#include <string>

#include "xdevice.h"
#include "ControllerDevice.h"
#include "TrackerDevice.h"
#include "Devices.h"
#include "Mathf.h"
#include "string_jni.h"
#include "blobApi.h"
#include "XCobraDevice.h"

extern void BLOB_CALL OnBlobAlgCallbackNoRecast(int num, int* id, float *pos);
extern void ControllerState_c2j(JNIEnv* env, jobject jdest,ControllerState* src);

#ifdef __cplusplus
extern "C"{
#endif

    JNIEXPORT jint JNICALL Java_com_ximmerse_natives_XDeviceDecoder_decodeControllerState_101(JNIEnv * env, jclass clazz,jbyteArray jdest, jbyteArray jbuffer, jint joffset, jint jcount){
        int size=sizeof(OneHandedControllerState_3DOF);
        OneHandedControllerState_3DOF state;
        memset(&state,0,size);
        //
        unsigned char* buffer=(unsigned char*)env->GetPrimitiveArrayCritical(jbuffer, NULL);
            Convert_ControllerStateFromBytes_01(&state,buffer,joffset,jcount);
        env->ReleasePrimitiveArrayCritical(jbuffer,buffer,0);
        buffer=(unsigned char*)env->GetPrimitiveArrayCritical(jdest, NULL);
            memcpy(buffer,&state,size);
        env->ReleasePrimitiveArrayCritical(jdest,buffer,0);
        return 0;
    }

    JNIEXPORT jint JNICALL Java_com_ximmerse_natives_XDeviceDecoder_decodeControllerState_102Auto(JNIEnv * env, jclass clazz, jint jwhich, jbyteArray jdest, jbyteArray jbuffer, jint joffset, jint jcount){
        int size=sizeof(OneHandedControllerState_3DOF);
        OneHandedControllerState_3DOF state;
        memset(&state,0,size);
        //
        unsigned char* buffer=(unsigned char*)env->GetPrimitiveArrayCritical(jbuffer, NULL);
            Convert_ControllerStateFromBytes_02Auto(XDeviceGetInputDevice(jwhich),&state,buffer,joffset,jcount);
        env->ReleasePrimitiveArrayCritical(jbuffer,buffer,0);
        buffer=(unsigned char*)env->GetPrimitiveArrayCritical(jdest, NULL);
            memcpy(buffer,&state,size);
        env->ReleasePrimitiveArrayCritical(jdest,buffer,0);
        return 0;
    }


    JNIEXPORT jint JNICALL Java_com_ximmerse_natives_XDeviceDecoder_decodeControllerState_103Daydream(JNIEnv * env, jclass clazz, jint jwhich, jbyteArray jdest, jbyteArray jbuffer, jint joffset, jint jcount){
     XCobraDevice*device  = (XCobraDevice*)XDeviceGetInputDevice(jwhich);
     if (device==NULL || device->inputState==NULL){return -1;}
        unsigned char* buffer=(unsigned char*)env->GetPrimitiveArrayCritical(jbuffer, NULL);
        Convert_ControllerStateFromDaydreamBytes(device->inputState,buffer,joffset,jcount);
        env->ReleasePrimitiveArrayCritical(jbuffer,buffer,0);
        ((XCobraDevice*)XDeviceGetInputDevice(jwhich))->DecorateInputState();
        return -1;
    }

    JNIEXPORT jint JNICALL Java_com_ximmerse_natives_XDeviceDecoder_decodeControllerState(JNIEnv * env, jclass clazz,jobject jdest, jint jwhich, jint jversion, jbyteArray jbuffer, jint joffset, jint jcount){
        InputDevice* thiz=XDeviceGetInputDevice(jwhich);
        if(thiz==NULL) return -1;
        //
        OneHandedControllerState_3DOF src={0};
        ControllerState inputState={0};
        //
        unsigned char* buffer=(unsigned char*)env->GetPrimitiveArrayCritical(jbuffer, NULL);
        switch(jversion){
            // TODO : Daydream Supports
            case 3:
                Convert_ControllerStateFromDaydreamBytes(&inputState,buffer,joffset,jcount);
            break;
            case 2:
                Convert_ControllerStateFromBytes_02Auto(thiz,&src,buffer,joffset,jcount);
                Convert_ControllerState(thiz,&inputState,&src);
            break;
            case 1:
                Convert_ControllerStateFromBytes_01(&src,buffer,joffset,jcount);
                Convert_ControllerState(thiz,&inputState,&src);
            break;
            default:
                // Not supported
            break;
        }
        env->ReleasePrimitiveArrayCritical(jbuffer,buffer,0);
        ControllerState_c2j(env,jdest,&inputState);
        return 0;
    }

#define ACTION_DOWN 0
#define ACTION_UP   1

#define USING_CONTROLLER_STATE_BEGIN() \
    InputDevice* device=XDeviceGetInputDevice(jwhich);             \
    if (device==NULL||device->type!=kDeviceType_ControllerDevice){ \
        return;                                                    \
    }                                                              \
    ControllerDevice* controller=(ControllerDevice*)device;        \
    ControllerState* state=controller->inputState;                 \
    controller->inputStateLock->lock()

#define USING_CONTROLLER_STATE_END() \
    controller->inputStateLock->unlock()

    JNIEXPORT void JNICALL Java_com_ximmerse_hardware_DeviceConnection_internalUpdateControllerTouchpad(JNIEnv * env,jclass clazz,jint jwhich,jint fingerId,jboolean isTouch,jfloat x,jfloat y){
        USING_CONTROLLER_STATE_BEGIN();
            if(fingerId==-1){
                state->axes[CONTROLLER_AXIS_LEFT_TRIGGER]=x;
                state->axes[CONTROLLER_AXIS_RIGHT_TRIGGER]=y;
            }else if(isTouch){
                state->axes[CONTROLLER_AXIS_LEFT_THUMB_X]=Mathf_Lerp(-1.0f,1.0f,Mathf_Normalize(0,1,x));
                state->axes[CONTROLLER_AXIS_LEFT_THUMB_Y]=Mathf_Lerp(1.0f,-1.0f,Mathf_Normalize(0,1,y));
                // For Daydream standard.
                state->axes[CONTROLLER_AXIS_RIGHT_THUMB_X]=state->axes[CONTROLLER_AXIS_LEFT_THUMB_X];
                state->axes[CONTROLLER_AXIS_RIGHT_THUMB_Y]=state->axes[CONTROLLER_AXIS_LEFT_THUMB_Y];
                //
                state->buttons|=CONTROLLER_BUTTON_LEFT_THUMB_MOVE;
            }else{
                state->axes[CONTROLLER_AXIS_LEFT_THUMB_X]=0.0f;
                state->axes[CONTROLLER_AXIS_LEFT_THUMB_Y]=0.0f;
                // For Daydream standard.
                //
                state->buttons&=~CONTROLLER_BUTTON_LEFT_THUMB_MOVE;
            }
        USING_CONTROLLER_STATE_END();
    }

    JNIEXPORT void JNICALL Java_com_ximmerse_hardware_DeviceConnection_internalUpdateControllerButtons(JNIEnv * env,jclass clazz,jint jwhich,jint jbuttons){
        USING_CONTROLLER_STATE_BEGIN();
            state->buttons=jbuttons;
        USING_CONTROLLER_STATE_END();
    }

    JNIEXPORT void JNICALL Java_com_ximmerse_hardware_DeviceConnection_internalUpdateControllerRotation(JNIEnv * env,jclass clazz,jint jwhich,jfloat x,jfloat y,jfloat z,jfloat w){
        InputDevice* device=XDeviceGetInputDevice(jwhich);
        if (device==NULL||device->type!=kDeviceType_ControllerDevice){
            return;
        }
        ControllerDevice* controller=(ControllerDevice*)device;
        float data[4];
        data[0]= x;
        data[1]= y;
        data[2]= z;
        data[3]= w;
        controller->UpdateRotation(0,true,data,false);
    }

    JNIEXPORT void JNICALL Java_com_ximmerse_hardware_DeviceConnection_internalUpdateControllerAccelerometer(JNIEnv * env,jclass clazz,jint jwhich,jfloat x,jfloat y,jfloat z){
        USING_CONTROLLER_STATE_BEGIN();
            int i=0;
            state->accelerometer[i++]=x;
            state->accelerometer[i++]=y;
            state->accelerometer[i++]=z;
        USING_CONTROLLER_STATE_END();
    }

    JNIEXPORT void JNICALL Java_com_ximmerse_hardware_DeviceConnection_internalUpdateControllerGyroscope(JNIEnv * env,jclass clazz,jint jwhich,jfloat x,jfloat y,jfloat z){
        USING_CONTROLLER_STATE_BEGIN();
            int i=0;
            state->gyroscope[i++]=x;
            state->gyroscope[i++]=y;
            state->gyroscope[i++]=z;
        USING_CONTROLLER_STATE_END();
    }

    JNIEXPORT void JNICALL Java_com_ximmerse_hardware_DeviceConnection_internalUpdateControllerState(JNIEnv * env,jclass clazz,jint jwhich){
        USING_CONTROLLER_STATE_BEGIN();
        USING_CONTROLLER_STATE_END();
        // TODO:
        ++state->timestamp;
        controller->DecorateInputState();
    }

#undef USING_CONTROLLER_STATE_BEGIN
#undef USING_CONTROLLER_STATE_END

#if ENABLE_XHAWK

    JNIEXPORT void JNICALL Java_com_ximmerse_hardware_DeviceConnection_internalUpdateTrackingResult(JNIEnv * env,jclass clazz,jint jwhich,jint jnum,jintArray jid,jfloatArray jdata){
//        InputDevice* device=XDeviceGetInputDevice(jwhich);
//        if(device==NULL||device->type!=kDeviceType_TrackerDevice){
//            return;
//        }
//        TrackerDevice* tracker=(TrackerDevice*)device;
        int* id=(int*)malloc(jnum*sizeof(int));
        float* data=(float*)malloc(jnum*3*sizeof(float));
            memcpy_j2c(env,id,jid,0,jnum*sizeof(int));
            memcpy_j2c(env,data,jdata,0,jnum*3*sizeof(float));
            //tracker->UpdateTrackingResult(jnum,id,data);
            OnBlobAlgCallbackNoRecast(jnum,id,data);
        free(id);
        free(data);
    }

    #include "blobApi.h"

    extern IMU_INFO g_CachedGetRawIMU;

    JNIEXPORT void JNICALL Java_com_ximmerse_hardware_DeviceConnection_internalUpdateImu(JNIEnv * env,jclass clazz,jint jwhich, jfloat jx, jfloat jy, jfloat jz, jfloat jw){
        ++g_CachedGetRawIMU.timestamp;
        g_CachedGetRawIMU.qx=jx;
        g_CachedGetRawIMU.qy=jy;
        g_CachedGetRawIMU.qz=jz;
        g_CachedGetRawIMU.qw=jw;

    }

#endif

#ifdef __cplusplus
}
#endif
