#include <jni.h>
#include "string_jni.h"
#include "hidapi.h"

//typedef jlong jobject_ptr;
#define jobject_ptr jlong

hid_device* hid_device_j2c(jobject_ptr jdevice) {
	return reinterpret_cast<hid_device*>(jdevice);
}

jobject_ptr hid_device_c2j(hid_device* device) {
	return reinterpret_cast<jobject_ptr>(device);
}


hid_device_info* hid_device_info_j2c(jobject_ptr jdevs) {
	return reinterpret_cast<hid_device_info*>(jdevs);
}

jobject_ptr hid_device_info_c2j(hid_device_info* devs) {
	return reinterpret_cast<jobject_ptr>(devs);
}

#ifdef __cplusplus
extern "C"{
#endif

	// Common

	JNIEXPORT jint JNICALL Java_com_ximmerse_natives_HidApi_hid_1init(JNIEnv * env, jclass clazz) {
		return hid_init();
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_natives_HidApi_hid_1exit(JNIEnv * env, jclass clazz) {
		return hid_exit();
	}


	// Enumerate

	JNIEXPORT jobject_ptr JNICALL Java_com_ximmerse_natives_HidApi_hid_1enumerate(JNIEnv * env, jclass clazz, jshort jvendor_id, jshort jproduct_id) {
		return hid_device_info_c2j(hid_enumerate(jvendor_id, jproduct_id));
	}

	JNIEXPORT void JNICALL Java_com_ximmerse_natives_HidApi_hid_1free_1enumeration(JNIEnv * env, jclass clazz, jobject_ptr jdevs) {
		hid_free_enumeration(hid_device_info_j2c(jdevs));
	}

	/* Android Ext */
	JNIEXPORT jobject_ptr JNICALL Java_com_ximmerse_natives_HidApi_hid_1enumerate_1next(JNIEnv * env, jclass clazz, jobject_ptr jdevs) {
		hid_device_info* devs=hid_device_info_j2c(jdevs);
		if(devs==NULL) {
			return 0;
		}
		return hid_device_info_c2j(devs->next);
	}

	// Open/Close

	JNIEXPORT jobject_ptr JNICALL Java_com_ximmerse_natives_HidApi_hid_1open(JNIEnv * env, jclass clazz, jshort jvendor_id, jshort jproduct_id, jstring jserial_number) {
		wchar_t* serial_number=js2w(env, jserial_number);
		jobject_ptr ret=hid_device_c2j(hid_open(jvendor_id, jproduct_id, serial_number));
		return ret;
	}

	JNIEXPORT jobject_ptr JNICALL Java_com_ximmerse_natives_HidApi_hid_1open_1path(JNIEnv * env, jclass clazz, jstring jpath) {
		char* path=js2c(env, jpath);
		jobject_ptr ret=hid_device_c2j(hid_open_path(path));
		return ret;
	}

	JNIEXPORT void JNICALL Java_com_ximmerse_natives_HidApi_hid_1close(JNIEnv * env, jclass clazz, jobject_ptr jdevice) {
		hid_close(hid_device_j2c(jdevice));
	}

	// Read/Write

	JNIEXPORT jint JNICALL Java_com_ximmerse_natives_HidApi_hid_1read(JNIEnv * env, jclass clazz, jobject_ptr jdevice, jbyteArray jdata, jint joffset, jint jlength){
		unsigned char* data=(unsigned char*)env->GetPrimitiveArrayCritical(jdata, NULL);
			int ret=hid_read(hid_device_j2c(jdevice),data+joffset,jlength);
		env->ReleasePrimitiveArrayCritical(jdata,data,0);
		return ret;
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_natives_HidApi_hid_1read_1timeout(JNIEnv * env, jclass clazz, jobject_ptr jdevice, jbyteArray jdata, jint joffset, jint jlength, jint jmilliseconds){
		unsigned char* data=(unsigned char*)env->GetPrimitiveArrayCritical(jdata, NULL);
			int ret=hid_read_timeout(hid_device_j2c(jdevice),data+joffset,jlength,jmilliseconds);
		env->ReleasePrimitiveArrayCritical(jdata,data,0);
		return ret;
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_natives_HidApi_hid_1get_1feature_1report(JNIEnv * env, jclass clazz, jobject_ptr jdevice, jbyteArray jdata, jint joffset, jint jlength){
		unsigned char* data=(unsigned char*)env->GetPrimitiveArrayCritical(jdata, NULL);
			int ret=hid_get_feature_report(hid_device_j2c(jdevice),data+joffset,jlength);
		env->ReleasePrimitiveArrayCritical(jdata,data,0);
		return ret;
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_natives_HidApi_hid_1write(JNIEnv * env, jclass clazz, jobject_ptr jdevice, jbyteArray jdata, jint joffset, jint jlength){
		unsigned char* data=(unsigned char*)env->GetPrimitiveArrayCritical(jdata, NULL);
			int ret=hid_write(hid_device_j2c(jdevice),data+joffset,jlength);
		env->ReleasePrimitiveArrayCritical(jdata,data,0);
		return ret;
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_natives_HidApi_hid_1send_1feature_1report(JNIEnv * env, jclass clazz, jobject_ptr jdevice, jbyteArray jdata, jint joffset, jint jlength){
		unsigned char* data=(unsigned char*)env->GetPrimitiveArrayCritical(jdata, NULL);
			int ret=hid_send_feature_report(hid_device_j2c(jdevice),data+joffset,jlength);
		env->ReleasePrimitiveArrayCritical(jdata,data,0);
		return ret;
	}

	// Properties

	/* Android Ext */
	JNIEXPORT jstring JNICALL Java_com_ximmerse_natives_HidApi_hid_1get_1path_1string(JNIEnv * env, jclass clazz, jobject_ptr jdevs) {
		hid_device_info* devs=hid_device_info_j2c(jdevs);
		if(devs==NULL) {
			return NULL;
		}
		jstring jpath=c2js(env,devs->path);
		return jpath;
	}

#ifdef __cplusplus
}
#endif
