#include "string_jni.h"

void memcpy_c2j(JNIEnv* env, jarray jdest, jint joffset, void* src, int size){
	unsigned char* dest=(unsigned char*)env->GetPrimitiveArrayCritical(jdest,NULL);
		memcpy(dest+joffset,src,size);
	env->ReleasePrimitiveArrayCritical(jdest,dest,0);
}

void memcpy_j2c(JNIEnv* env, void* dest, jarray jsrc, jint joffset, int size){
	unsigned char* src=(unsigned char*)env->GetPrimitiveArrayCritical(jsrc,NULL);
		memcpy(dest,src+joffset,size);
	env->ReleasePrimitiveArrayCritical(jsrc,src,0);
}

void* GetSafeJniBuffer(JNIEnv* env, jarray jsrc, jint joffset, jint jcount, jint mode){
	void* buffer=malloc(jcount);
	memcpy_j2c(env,buffer,jsrc,joffset,jcount);
	return buffer;
}

void ReleaseSafeJniBuffer(JNIEnv* env, jarray jsrc, jint joffset, jint jcount, void* buffer, jint mode){
	memcpy_c2j(env,jsrc,joffset,buffer,jcount);
	free(buffer);
}

// http://blog.sina.com.cn/s/blog_4119bd830100t4ap.html

jstring c2js(JNIEnv* env, char* str){
	jclass strClass = env->FindClass("java/lang/String");
	jmethodID ctorID = env->GetMethodID(strClass, "<init>", "([BLjava/lang/String;)V");
	jbyteArray bytes = env->NewByteArray(strlen(str));
	env->SetByteArrayRegion(bytes, 0, strlen(str), (jbyte*)str);
	jstring encoding = env->NewStringUTF("utf-8");
	return (jstring)env->NewObject(strClass, ctorID, bytes, encoding);
}

char* js2c(JNIEnv* env, jstring jstr){
	char* rtn = NULL;
	jclass clsstring = env->FindClass("java/lang/String");
	jstring strencode = env->NewStringUTF("utf-8");
	jmethodID mid = env->GetMethodID(clsstring, "getBytes", "(Ljava/lang/String;)[B");
	jbyteArray barr = (jbyteArray)env->CallObjectMethod(jstr, mid, strencode);
	jsize alen = env->GetArrayLength(barr);
	jbyte* ba = env->GetByteArrayElements(barr, JNI_FALSE);
	if (alen > 0)
	{
		rtn = (char*)malloc(alen + 1);
		memcpy(rtn, ba, alen);
		rtn[alen] = 0;
	}
	env->ReleaseByteArrayElements(barr, ba, 0);
	return rtn;
}

jstring w2js(JNIEnv* env, wchar_t* wstr){
	size_t len = wcslen(wstr);
	jchar* str2 = (jchar*)malloc(sizeof(jchar)*(len + 1));
	int i;
	for (i = 0; i < len; i++)
		str2[i] = wstr[i];
	str2[len] = 0;
	jstring js = env->NewString(str2, len);
	free(str2);
	return js;
}

wchar_t* js2w(JNIEnv* env, jstring jstr){
	int len = env->GetStringLength(jstr);
	wchar_t *w_buffer = new wchar_t[len];
	memset(w_buffer, 0, len + 1);
	w_buffer[len] = '\0';
	wcsncpy(w_buffer, (wchar_t *)env->GetStringChars(jstr, 0), len);
	env->ReleaseStringChars(jstr, (const unsigned short *)w_buffer);
	return w_buffer;
}
