#include "xdevice_includes.h"
#include <jni.h>
#include "string_jni.h"

void ControllerState_c2j(JNIEnv* env, jobject jdest,ControllerState* src) {
	jclass clazz=env->GetObjectClass(jdest);
		jfieldID field=NULL;
		// int
		field=env->GetFieldID(clazz,"handle","I");if(field){env->SetIntField(jdest,field,src->handle);}
		field=env->GetFieldID(clazz,"timestamp","I");if(field){env->SetIntField(jdest,field,src->timestamp);}
		field=env->GetFieldID(clazz,"state_mask","I");if(field){env->SetIntField(jdest,field,src->state_mask);}
		field=env->GetFieldID(clazz,"buttons","I");if(field){env->SetIntField(jdest,field,src->buttons);}
		// float[]
		field=env->GetFieldID(clazz,"axes","[F");if(field){memcpy_c2j(env,(jarray)env->GetObjectField(jdest,field),0,src->axes,sizeof(float)*6);}
		field=env->GetFieldID(clazz,"position","[F");if(field){memcpy_c2j(env,(jarray)env->GetObjectField(jdest,field),0,src->position,sizeof(float)*3);}
		field=env->GetFieldID(clazz,"accelerometer","[F");if(field){memcpy_c2j(env,(jarray)env->GetObjectField(jdest,field),0,src->accelerometer,sizeof(float)*3);}
		field=env->GetFieldID(clazz,"rotation","[F");if(field){memcpy_c2j(env,(jarray)env->GetObjectField(jdest,field),0,src->rotation,sizeof(float)*4);}
		field=env->GetFieldID(clazz,"gyroscope","[F");if(field){memcpy_c2j(env,(jarray)env->GetObjectField(jdest,field),0,src->gyroscope,sizeof(float)*3);}
	env->DeleteLocalRef(clazz);
}

void TrackerState_c2j(JNIEnv* env, jobject jdest,TrackerState* src) {
	jclass clazz=env->GetObjectClass(jdest);
		jfieldID field=NULL;
		// int
		field=env->GetFieldID(clazz,"handle","I");if(field){env->SetIntField(jdest,field,src->handle);}
		field=env->GetFieldID(clazz,"frameCount","I");if(field){env->SetIntField(jdest,field,src->frameCount);}
		field=env->GetFieldID(clazz,"timestamp","I");if(field){env->SetIntField(jdest,field,src->timestamp);}
		field=env->GetFieldID(clazz,"capacity","I");if(field){env->SetIntField(jdest,field,src->capacity);}
		field=env->GetFieldID(clazz,"count","I");if(field){env->SetIntField(jdest,field,src->count);}
		// float[]
		field=env->GetFieldID(clazz,"id","[I");if(field){memcpy_c2j(env,(jarray)env->GetObjectField(jdest,field),0,src->id,sizeof(int)*src->count);}
		field=env->GetFieldID(clazz,"data","[F");if(field){memcpy_c2j(env,(jarray)env->GetObjectField(jdest,field),0,src->data,sizeof(float)*3*src->count);}
	env->DeleteLocalRef(clazz);
}

extern DeviceContext* sDeviceContext;

void initJniContext(JNIEnv* env,jclass clazz){
	env->GetJavaVM(&sDeviceContext->javaVM);
	sDeviceContext->clazz=(jclass)env->NewGlobalRef(env->FindClass("com/ximmerse/sdk/XDeviceApi"));
}

void exitJniContext(JNIEnv* env,jclass clazz){
	if(sDeviceContext->clazz!=NULL){
		env->DeleteGlobalRef(sDeviceContext->clazz);
		sDeviceContext->clazz=NULL;
	}
	sDeviceContext->ClearJniCaches();
}

#ifdef __cplusplus
extern "C" {
#endif

	// Common

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalInit(JNIEnv* env, jclass clazz) {
		jint ret=XDeviceInit();
		initJniContext(env,clazz);
		return ret;
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalExit(JNIEnv* env, jclass clazz) {
		exitJniContext(env,clazz);
		return XDeviceExit();
	}

	// Device Management

	JNIEXPORT jlong JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalGetContext(JNIEnv* env, jclass clazz,jboolean jautoCreate) {
		DeviceContext* ret=XDeviceGetContext(jautoCreate);
		if(ret==NULL){return 0;}
		else{return reinterpret_cast<jlong>(ret);}
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalGetInputDeviceHandle(JNIEnv* env, jclass clazz, jstring jname) {
		char* name=js2c(env, jname);
		jint ret=XDeviceGetInputDeviceHandle(name);
		return ret;
	}

	JNIEXPORT jstring JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalGetInputDeviceName(JNIEnv* env, jclass clazz, jint jwhich) {
		return c2js(env,XDeviceGetInputDeviceName(jwhich));
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalGetInputDeviceCount(JNIEnv* env, jclass clazz) {
		return XDeviceGetInputDeviceCount();
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalGetInputDevices(JNIEnv* env, jclass clazz, jint jtype, jintArray jwhichBuffer,jint jwhichBufferSize) {
		int* whichBuffer=(int*)env->GetPrimitiveArrayCritical(jwhichBuffer,NULL);
			int ret=XDeviceGetInputDevices(jtype,whichBuffer,jwhichBufferSize);
		env->ReleasePrimitiveArrayCritical(jwhichBuffer,whichBuffer,0);
		return ret;
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalRemoveInputDeviceAt(JNIEnv* env, jclass clazz, jint jwhich, jboolean jdispose) {
		InputDevice* ret=XDeviceRemoveInputDeviceAt(jwhich,jdispose);
		if(ret==NULL){return 0;}
		else{return reinterpret_cast<jlong>(ret);}
	}

	// I/O

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalGetInputState(JNIEnv* env, jclass clazz, jint jwhich, jobject jstate) {
		ControllerState state;
		jint ret=XDeviceGetInputState(jwhich, &state);
		if(ret==0) {
			ControllerState_c2j(env, jstate, &state);
		}
		return ret;
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalGetInputState1(JNIEnv* env, jclass clazz, jint jwhich, jobject jstate) {
		TrackerState state;
		jint ret=XDeviceGetInputState(jwhich, &state);
		if(ret==0) {
			TrackerState_c2j(env, jstate, &state);
		}
		return ret;
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalDispatchEventListener(JNIEnv* env, jclass clazz, jint jwhich, jint jeventID) {
		return XDeviceDispatchEventListener(jwhich, jeventID);
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalSendMessage(JNIEnv* env, jclass clazz, jint jwhich, jint jMsg, jint jwParam, jint jlParam) {
		return XDeviceSendMessage(jwhich, jMsg, jwParam, jlParam);
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalResetInputState(JNIEnv* env, jclass clazz, jint jwhich) {
		return XDeviceResetInputState(jwhich);
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalUpdateInputState(JNIEnv* env, jclass clazz, jint jwhich) {
		return XDeviceUpdateInputState(jwhich);
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalUpdateInputBuffer(JNIEnv* env, jclass clazz, jint jwhich, jarray jbuffer, jint joffset, jint jcount) {
		unsigned char* buffer=(unsigned char*)GetSafeJniBuffer(env,jbuffer,joffset,jcount,0);
			jint ret=XDeviceUpdateInputBuffer(jwhich,buffer,0,jcount);
		ReleaseSafeJniBuffer(env,jbuffer,joffset,jcount,buffer,0);
		return ret;
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalSerializeMessage(JNIEnv* env, jclass clazz, jarray jbuffer, jint joffset, jint jcount, jint jwhich, jint jMsg, jint jwParam, jint jlParam) {
		unsigned char* buffer = (unsigned char*)env->GetPrimitiveArrayCritical(jbuffer, NULL);
			jint ret=XDeviceSerializeMessage(buffer,joffset,jcount,jwhich,jMsg,jwParam,jlParam);
		env->ReleasePrimitiveArrayCritical(jbuffer, buffer, 0);
		return ret;
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalSendBuffer(JNIEnv* env, jclass clazz, jint jwhich, jarray jbuffer, jint joffset, jint jcount) {
		unsigned char* buffer = (unsigned char*)env->GetPrimitiveArrayCritical(jbuffer, NULL);
			jint ret=XDeviceSendBuffer(jwhich,buffer,joffset,jcount);
		env->ReleasePrimitiveArrayCritical(jbuffer, buffer, 0);
		return ret;
	}

	// Get/Set

	JNIEXPORT jboolean JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalGetBool(JNIEnv* env, jclass clazz, jint jwhich, jint jfieldID, jboolean jdefaultValue) {
		return XDeviceGetBool(jwhich,jfieldID,jdefaultValue);
	}

	JNIEXPORT void JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalSetBool(JNIEnv* env, jclass clazz, jint jwhich, jint jfieldID, jboolean jvalue) {
		XDeviceSetBool(jwhich,jfieldID,jvalue);
	}

	JNIEXPORT jint JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalGetInt(JNIEnv* env, jclass clazz, jint jwhich, jint jfieldID, jint jdefaultValue) {
		return XDeviceGetInt(jwhich,jfieldID,jdefaultValue);
	}

	JNIEXPORT void JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalSetInt(JNIEnv* env, jclass clazz, jint jwhich, jint jfieldID, jint jvalue) {
		XDeviceSetInt(jwhich,jfieldID,jvalue);
	}

	JNIEXPORT jfloat JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalGetFloat(JNIEnv* env, jclass clazz, jint jwhich, jint jfieldID, jfloat jdefaultValue) {
		return XDeviceGetFloat(jwhich,jfieldID,jdefaultValue);
	}

	JNIEXPORT void JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalSetFloat(JNIEnv* env, jclass clazz, jint jwhich, jint jfieldID, jfloat jvalue) {
		XDeviceSetFloat(jwhich,jfieldID,jvalue);
	}

	JNIEXPORT void JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalGetObject(JNIEnv* env, jclass clazz, jint jwhich, jint jfieldID, jarray jbuffer, jint joffset, jint jcount) {
		unsigned char* buffer = (unsigned char*)env->GetPrimitiveArrayCritical(jbuffer, NULL);
			XDeviceGetObject(jwhich,jfieldID,buffer,joffset,jcount);
		env->ReleasePrimitiveArrayCritical(jbuffer, buffer, 0);
	}

	JNIEXPORT void JNICALL Java_com_ximmerse_sdk_NativeXDeviceApi_internalSetObject(JNIEnv* env, jclass clazz, jint jwhich, jint jfieldID, jarray jbuffer, jint joffset, jint jcount) {
		unsigned char* buffer = (unsigned char*)env->GetPrimitiveArrayCritical(jbuffer, NULL);
			XDeviceSetObject(jwhich,jfieldID,buffer,joffset,jcount);
		env->ReleasePrimitiveArrayCritical(jbuffer, buffer, 0);
	}

#ifdef __cplusplus
}
#endif
